local require = require

local unistd = require("posix.unistd")
local system = require("common.system")
local cjson = require("cjson.safe")
local pystring = require("pystring")
local ChaElfSyms = require("haWork.haElfSyms")
local haPidMap = require("haWork.haPidMap")
local liveTrace = require("liveTrace.liveTrace")

local pcall = pcall
local getpid = unistd.getpid
local unpack = unpack
local open = io.open
local jencode = cjson.encode
local jdecode = cjson.decode
local fomart = string.format
local match = string.match
local split = pystring.split
local startswith = pystring.startswith
local genMap = haPidMap.genMap
local getMaps = liveTrace.getMaps
local getWorkList = liveTrace.getWorList
local sysconf = unistd.sysconf

local haSyms = ChaElfSyms.new()

local mt = {}

local function nproc(_)
    return {nporc = sysconf(84)}
end

local function selfPid(_)
    return {pid = getpid()}
end

local function kallSyms(_)
    return haSyms:getElfs(nil, "/proc/kallsyms")
end

local function pidExist(args)
    local path = "/proc/" .. args.pid .. "/comm"
    local f, _ = open(path, 'r')
    if f then
        local comm = f:read()
        f:close()
        return {path = path, comm = comm}
    end
    return {path = false, comm = false}
end

local function pidMapSid(args)
    local pid = args.pid
    local ok, maps

    ok, maps = pcall(genMap, pid)

    if ok then
        local f, sid, path
        path = "/proc/" .. pid .. "/cgroup"
        f = open(path, 'r')
        if f then
            sid = "host"
            for line in f:lines() do
                local s
                _, s = unpack(split(line, ":", 1))
                if startswith(s, "name=systemd") then
                    sid = match(s, "%x%x%x%x%x%x%x%x%x%x%x%x%x%x%x%x%x%x%x%x%x%x%x%x")
                    if sid then
                        sid = sid:sub(1, 12)
                    else
                        sid = "host"
                    end
                    break
                end
            end
            f:close()
            return {maps = maps, sid = sid}
        end
    end
    return {}
end

local function elfSym(args)
    return haSyms:getElfs(args.pid, args.elf)
end

local function liveDump(args)
    local res = getMaps(args.module)
    return res
end

local function liveWork(_)
    return getWorkList()
end

local apis = {
    nproc = nproc,
    selfPid = selfPid,
    kallSyms = kallSyms,
    pidExist = pidExist,
    pidMapSid = pidMapSid,
    elfSym = elfSym,
    liveDump = liveDump,
    liveWork = liveWork,
}

function mt.api(tReq)
    local res = jdecode(tReq.body)
    if not res then
        return {
            code = 400,
            body = "invalid request"
        }
    end

    if apis[res.api] then
        return {
            code = 200,
            body = jencode(apis[res.api](res.args)) or ""
        }
    end
    return {
        code = 400,
        body = fomart("invalid api unit %s", res.api)
    }
end

return mt
